/*
 * This code resource analytically determines the maxima and minima of a spline. It
 * does so by evaluating expressions for the x position that are a solution of the
 * equation where the first derivative is set to zero. Subsequently it checks the
 * second derivative so as to find out whether the found extremum is a maximum or a
 * minimum.
 * Maxima and minima on the outer two splinal pieces are ignored so that extrema
 * too near the edge are not considered. Potential numerical instability is avoided
 * by chosing a series expansion of the a,b,c formula when appropriate.
 *
 * This code resource is a building block for a peak detection procedure consisting
 * of a spline fit, this algorithm, noise determination and a SNR threshold filter.
 */

#include "extcode.h"
#include "hosttype.h"

/* stubs for advanced CIN functions */

UseDefaultCINInit
UseDefaultCINDispose
UseDefaultCINAbort
UseDefaultCINLoad
UseDefaultCINUnload
UseDefaultCINSave

/*
 * typedefs
 */

typedef struct {
	int32 dimSize;
	float64 data[1];
	} TD1;
typedef TD1 **TD1Hdl;

/*
 * Mark this CIN as reentrant.
 */
CIN MgErr CINProperties(int32 mode, void *data)
{
  switch (mode)
    {
      case kCINIsReentrant:
      *(Bool32 *)data = TRUE;
      return noErr;
      break;
    }
  return mgNotSupported;
}

CIN MgErr CINRun(TD1Hdl Xh, TD1Hdl Yh, TD1Hdl Rh, TD1Hdl maximaxh, TD1Hdl maximayh, TD1Hdl minimaxh, TD1Hdl minimayh);

CIN MgErr CINRun(TD1Hdl Xh, TD1Hdl Yh, TD1Hdl Rh, TD1Hdl maximaxh, TD1Hdl maximayh, TD1Hdl minimaxh, TD1Hdl minimayh)
{
  float64 *X = (*Xh)->data;	/* Spline joints X partitioning */
  float64 *Y = (*Yh)->data;	/* Spline Y values to be fit in gliding FITDIMS chunks */
  float64 *R = (*Rh)->data;	/* Second derivatives for gliding and final storage */
  float64 *maximax;
  float64 *maximay;
  float64 *minimax;
  float64 *minimay;

  int32 joints = (*Xh)->dimSize;
  int32 joint, maxis, minis;

  /* Check if the X, Y and R arrays have the same size */
  if (joints != (*Yh)->dimSize || joints != (*Rh)->dimSize)
    {
      DbgPrintf("extrema.c: size mismatch between X, Y and 2nd derivative arrays!\n");
      return bogusError;
    }

  /* Check if we got a reasonable number of joints */
  if (joints < 4)
    {
      DbgPrintf("extrema.c: too few joints (<4) in spline!\n");
      return bogusError;
    }

  /* Resize the maxima and minima arrays so that they can hold maximum amount of expected extrema */
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)maximaxh, 2*joints))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*maximaxh)->dimSize = 2*joints;
  maximax = (*maximaxh)->data;
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)maximayh, 2*joints))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*maximayh)->dimSize = 2*joints;
  maximay = (*maximayh)->data;
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)minimaxh, 2*joints))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*minimaxh)->dimSize = 2*joints;
  minimax = (*minimaxh)->data;
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)minimayh, 2*joints))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*minimayh)->dimSize = 2*joints;
  minimay = (*minimayh)->data;

  /* Check if the X values increase monotonously */
  for (joint = 1; joint < joints; joint++)
    {
      if (X[joint-1] >= X[joint])
        {
          DbgPrintf("extrema.c: spline joint positions X must increase monotonously!\n");
          return bogusError;
        }
    }

  /*
   * This is the loop that calculates the zero derivative solutions and their applicability;
   * the zeroes of the first derivative of each splinal piece are evaluated. To be eligible
   * a zero must be real and it must lie in the x range of the splinal piece.
   */
  maxis = 0; minis = 0;
  for (joint = 1; joint < joints-2; joint++)
    {
      int32 j = joint;
      int32 k = joint+1;
      float64 h = X[k]-X[j];
      float64 Rd = R[k]-R[j];
      float64 preroot;
      /* First calculate the sum of terms under the root */
      preroot = h*h*(R[j]*R[j]+R[k]*R[k]+R[j]*R[k])-6*Rd*(Y[k]-Y[j]);

      /* Ignore complex solutions or solutions at zero second derivative (bending point plateau) */
      if (preroot > 0)
        {
          float64 signum, A, B, a, b, c, d, x, y;

          /*
           * Calculate the solution for A. Beware that when improperly sequenced the expression
           * has extreme numerical instability w.r.t. Rd=Rk-Rj. It is quite legal to have
           * Rk=Rj or nearly so. Because the expression is divided by Rd, large numbers will
           * be added or subtracted if this is not done as the last step. When Rk=Rj there
           * will be a division by 0 if not handled in a special way.
           * To see how things work out it is best to make a Taylor expansion of the a,b,c
           * formula (general solution for zeros of a parabola). It turns out that when
           * d = 4ac/b^2 << 1 numerical instability can occur both due the subtraction of two
           * large numbers as well as due to the division by a (which can be zero or almost so).
           * For the derivative w.r.t. A we have:
           * a = (R[k]-R[j])*h*h/2
           * b = R[j]*h*h
           * c = Y[k]-Y[j]-(2*R[j]+R[k])*h*h/6
           */
          a = Rd*h*h/2.0;
          b = R[j]*h*h;
          c = Y[k]-Y[j]-(2*R[j]+R[k])*h*h/6.0;
          d = 4*a*c/(b*b);

          /* Evaluate both roots (the derivative of a cubic splinal piece is a parabola */
          for (signum = -1; signum <= 1; signum += 2)
            {
              /* Select the numerically most stable expression for the current ratio d */
              if ( (signum*b > 0) && (fabs(d) < 0.05) )
                {
                  A = (-2*c/b)*(0.5+0.125*d+0.0625*d*d+0.0390625*d*d*d+0.02734375*d*d*d*d+0.0205078125*d*d*d*d*d+0.01611328125*d*d*d*d*d*d);
                }
              else
                {
                  A = (-R[j]+signum*sqrt(preroot)/(sqrt(3)*h))/Rd;
                }

              /* If the root lies within the splinal piece we've got an extremum */
              if (A >= 0 && A < 1)
                {
                  B = 1-A;

                  /* Calculate the x position of the extremum */
                  x = A*X[k]+B*X[j];

                  /* Calculate the y position of the extremum on the splinal piece */
                  y = A*Y[k]+B*Y[j]+h*h*((A*A*A-A)*R[k]+(B*B*B-B)*R[j])/6;

                  /*
                   * Interpolate the 2nd derivative. If it is < 0 we've got a maximum, else it
                   * is a minimum. Store the x and y positions of the found extremum in the
                   * proper arrays.
                   */
                  if (A*R[k]+B*R[j] < 0)
                    {
                      maximax[maxis] = x;
                      maximay[maxis] = y;
                      maxis += 1;
                    }
                  else
                    {
                      minimax[minis] = x;
                      minimay[minis] = y;
                      minis += 1;
                    }
                }
            }
        }
    }

  /* Resize the maxima and minima arrays to the number of these actually found */
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)maximaxh, maxis))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*maximaxh)->dimSize = maxis;
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)maximayh, maxis))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*maximayh)->dimSize = maxis;
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)minimaxh, minis))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*minimaxh)->dimSize = minis;
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)minimayh, minis))
    {
      DbgPrintf("extrema.c: out of memory!");
      return bogusError;
    }
  (*minimayh)->dimSize = minis;

  return noErr;
}
