/*
 * This code resource performs a batch interpolation of a cubic spline defined through
 * the X and Y values of its joints and redundantly defined (for efficient evaluation)
 * through its second derivates at those joints.
 *
 * Batch interpolation means that a series of x values is in one go as opposed to the
 * "Spline Interpolation" analysis VI which can only process one x value at a time and
 * thus is less efficient if a bunch of x values need to be interpolated in one go.
 *
 * It is required that both the x and X arrays are monotonously increasing. Parts of
 * the x array that are outside of the range covered by the X spline joint positions
 * are extrapolated from the near outermost splinal piece.
 */

#include "extcode.h"
#include "hosttype.h"

/* stubs for advanced CIN functions */

UseDefaultCINInit
UseDefaultCINDispose
UseDefaultCINAbort
UseDefaultCINLoad
UseDefaultCINUnload
UseDefaultCINSave

/*
 * typedefs
 */

typedef struct {
	int32 dimSize;
	float64 data[1];
	} TD1;
typedef TD1 **TD1Hdl;

/*
 * Mark this CIN as reentrant.
 */
CIN MgErr CINProperties(int32 mode, void *data)
{
  switch (mode)
    {
      case kCINIsReentrant:
      *(Bool32 *)data = TRUE;
      return noErr;
      break;
    }
  return mgNotSupported;
}

CIN MgErr CINRun(TD1Hdl xh, TD1Hdl ph, TD1Hdl Xh, TD1Hdl Yh, TD1Hdl Rh);

CIN MgErr CINRun(TD1Hdl xh, TD1Hdl ph, TD1Hdl Xh, TD1Hdl Yh, TD1Hdl Rh)
{
  float64 *x = (*xh)->data;	/* The x coordinates at which to interpolate, x[i+1] > x[i] */
  float64 *p;			/* The splinal interpolates at the data points x */

  float64 *X = (*Xh)->data;	/* Spline joints X partitioning */
  float64 *Y = (*Yh)->data;	/* Spline Y values to be fit in gliding FITDIMS chunks */
  float64 *R = (*Rh)->data;	/* Second derivatives for gliding and final storage */

  int32 points = (*xh)->dimSize;
  int32 joints = (*Xh)->dimSize;
  int32 point, joint;

  /* Check if the X, Y and R arrays have the same size */
  if (joints != (*Yh)->dimSize || joints != (*Rh)->dimSize)
    {
      DbgPrintf("interpol.c: size mismatch between X, Y and 2nd derivative arrays!\n");
      return bogusError;
    }


  /* Check if we got a reasonable number of joints */
  if (joints < 2)
    {
      DbgPrintf("interpol.c: too few joints (<2) in spline, cannot interpolate!\n");
      return bogusError;
    }

  /* Resize the interpolates array so that it can hold the interpolation values and dereference it */
  if (noErr != NumericArrayResize(fD, 1, &(UHandle)ph, points))
    {
      DbgPrintf("spline.c: out of memory!");
      return bogusError;
    }
  (*ph)->dimSize = points;
  p = (*ph)->data;

  /* Check if the X values increase monotonously */
  for (joint = 1; joint < joints; joint++)
    {
      if (X[joint-1] >= X[joint])
        {
          DbgPrintf("interpol.c: spline joint positions X must increase monotonously!\n");
          return bogusError;
        }
    }

  /* Check if the x values increase monotonously */
  for (point = 1; point < points; point++)
    {
      if (x[point-1] >= x[point])
        {
          DbgPrintf("interpol.c: x points must increase monotonously!\n");
          return bogusError;
        }
    }

  /*
   * This is the actual interpolation loop
   * If x[0] < X[0] or x[points-1] > X[joints-1] we extrapolate the bits sticking out
   */
  {
    int32 j = 0;
    int32 k = 1;
    float64 h = X[k]-X[j];
    float64 A, B, AA, BB;
  
    for (point = 0; point < points; point++)
      {
        /* Search for the spline covering the regions within which x[point] falls */
        while (x[point] > X[k] && k < joints-1)
          {
            j += 1;
            k += 1;
            h = X[k]-X[j];
          }
  
        /* Efficiently and stably calculate the interpolation values */
        A = (x[point]-X[j])/h;
        B = 1-A;
        AA = A*A*A-A;
        BB = B*B*B-B;
        p[point] = B*Y[j]+A*Y[k]+(BB*R[j]+AA*R[k])*h*h/6;
      }
  }

  return noErr;
}
