/*
 * This code resource calculates a moderately accuracte measure for the noise in a trace.
 * It assumes that the features recorded in the trace are at least a couple of samples
 * wide so that any fluctuations between adjacent samples can be taken as a measure for
 * the noise.
 * The determined noise measure is weighted such that it approximately equals the rms
 * amplitude when supplied with a gaussian noise pattern.
 * This code resource is guaranteed to return a non zero value unless the trace is
 * perfectly constant; when no local fluctuations are found in the trace half the minimum
 * nonzero step between subsequent data points is returned. When the trace is perfectly
 * flat an erroneous result is returned.
 */

#include "extcode.h"
#include "hosttype.h"

/* stubs for advanced CIN functions */

UseDefaultCINInit
UseDefaultCINDispose
UseDefaultCINAbort
UseDefaultCINLoad
UseDefaultCINUnload
UseDefaultCINSave

/*
 * typedefs
 */

typedef struct {
	int32 dimSize;
	float32 data[1];
	} TD1;
typedef TD1 **TD1Hdl;

/*
 * Mark this CIN as reentrant.
 */
CIN MgErr CINProperties(int32 mode, void *data)
{
  switch (mode)
    {
      case kCINIsReentrant:
      *(Bool32 *)data = TRUE;
      return noErr;
      break;
    }
  return mgNotSupported;
}

CIN MgErr CINRun(TD1Hdl valuesh, float64 *noisep);

CIN MgErr CINRun(TD1Hdl valuesh, float64 *noisep)
{
  int32 i, noises;
  float64 averderi;
  float32 noise, uppernoisebound;

  /* Dereference and allocate */
  int32  points = (*valuesh)->dimSize;
  float32 *vals = (*valuesh)->data;
  float32 *deri = (float32 *)DSNewPClr(points*sizeof(float32));    /* Non-normalized numerical derivative */
  if (deri == NULL)
    {
      DbgPrintf("noise.c: out of memory!");
      return bogusError;
    }

  /* Check if we have atleast a semi-sensible amount of points */
  if (points < 50)
    {
      DbgPrintf("noise.c: too few points (%ld < 50), can not do noise statistics!\n", points);
      return bogusError;
    }

  /* Calculate the derivative and the average absolute derivative */
  averderi = 0;
  for (i = 0; i < points-1; i++)
    {
      deri[i] = vals[i+1]-vals[i];
      averderi += fabs(deri[i]);
    }
  averderi /= (points-1);

  /* Calculate a measure for the noise from wiggly (up down up) baseline corrected bits */
  noises = 0; noise = 0; uppernoisebound = fabs(deri[0])+1;
  for (i = 1; i < points-2; i++)
    {
      float32 avder, left, middle, right;

      avder = (deri[i-1]+deri[i]+deri[i+1])/3;
      left = deri[i-1]-avder;
      middle = deri[i]-avder;
      right = deri[i+1]-avder;

      /* If left and right have the same sign */
      if (left*right > 0)
        {
          /* But the sign is opposite to that of the middle difference */
          if (middle*right < 0)
            {
              /* And the derivative may not be too large (to avoid flanks) */
              if (fabs(deri[i]) < 2*averderi)
                {
                  /* Add wiggle to average sum */
                  noise += fabs(middle);
                  noises += 1;
                }
            }
        }
      /* As a failsafe, calculate an upper bound for the noise (half the smallest nonzero step) */
      if (fabs(deri[i]) < uppernoisebound && deri[i] != 0) uppernoisebound = fabs(deri[i]);
    }
  if (noises)
      noise = 0.677*noise/noises;        /* fudge factor makes it come out OK for Gaussian white noise */
  else
      noise = uppernoisebound/2;

  /* Return th result */
  *noisep = noise;

  /* Free allocated working arrays */
  DSDisposePtr(deri);

  /* All went OK, drop off the edge */
  return noErr;
}
