/*
 * This code resource fits a multipart cubic spline defined by the points (X,Y) to
 * noisy data (x,y). When not specified, the X partitioning is computed from the
 * X0 and deltaX values. The Y values are fit to a least squares difference between
 * the spline and the data.
 *
 * The fit is not global; instead of fitting the whole spline at once, the spline
 * is fit locally, FITDIMS pieces at a time while stepping from the lowest to highest
 * X value, one point at a time. This reduces the number of dimensions and allows
 * the fit to converge in a reasonable amount of time. This VI performs only one such
 * fitting step per call. During the first few calls state is initialized;
 * The p array is sized to the dimension of the x and y arrays.
 * The X array is created based on X0 and X when its dimension is zero.
 * The Y array is sized to the dimension of the X array.
 * The R array is sized to the dimension of the X array.
 * Some internal arrays are also initialized.
 * The CIN is made aware of which iteration it is in via the iteration input. It
 * notifies the calling loop of completion via the busy output. It is required
 * to loop back the p, X, Y and R arrays from the output to the input since
 * these are initialized only during the first iteration. The reason for this
 * approach is that a CIN is not interruptable, and the fitting might take a
 * long time. Putting the main loop outside of the CIN will therefore allow for
 * periodic interruptions.
 *
 * The deltaX value determines the size of the splinal pieces. It functions like
 * a sort of low pass filter cut-off; the pieces should be less wide than the
 * width of the narrowest feature size which the spline must be able to represent.
 * Natural boundary conditions (second derivative forced to zero) are choosen on
 * both ends of the spline.
 * The result is a collection of X points, the fit Y values and the second
 * derivatives R. Using these three arrays, spline interpolation can be performed.
 * Also the spline evaluated at the points x is returned in the p array.
 */

#include "extcode.h"
#include "hosttype.h"

#define FITDIMS 5
#define MAXITERATIONS 100
#define NATURAL 1e30

/* Some utility function declarations */

float64 minimize(float64 *x, float64 *y, float64 *p, int32 points, float64 *X, float64 *Y, float64 D, int32 trail);
void spline(float64 *R, float64 *X, float64 *Y, float64 left, float64 right, int32 joints);
void interpolate(float64 *p, float64 *x, int32 firstpoint, int32 lastpoint, float64 *X, float64 *Y, float64 *R, int32 joints);
float64 regres(float64 *xval, float64 *yval, float64 X, int32 elements);
void DisposeIfNN(float64 *allocedptr);

/* The next few arrays are needed by spline(), they are preallocated with JOINTS elements */
static float64 *alpha;
static float64 *h;
static float64 *l;
static float64 *mu;
static float64 *tri;
/* The following arrays are for efficient repeated interpolate() call, they are preallocated with POINTS elements */
static float64 *A, *B, *AA, *BB;

/* stubs for advanced CIN functions */

UseDefaultCINInit
UseDefaultCINDispose
UseDefaultCINAbort
UseDefaultCINLoad
UseDefaultCINUnload
UseDefaultCINSave

/*
 * typedefs
 */

typedef struct {
	int32 dimSize;
	float64 data[1];
	} TD1;
typedef TD1 **TD1Hdl;

/*
 * Mark this CIN as reentrant.
 * There is static data so it aint really, but access is serialized by the wrapping VI.
 */
CIN MgErr CINProperties(int32 mode, void *data)
{
  switch (mode)
    {
      case kCINIsReentrant:
      *(Bool32 *)data = TRUE;
      return noErr;
      break;
    }
  return mgNotSupported;
}

CIN MgErr CINRun(TD1Hdl xh, TD1Hdl yh, TD1Hdl ph, TD1Hdl Xh, TD1Hdl Yh, TD1Hdl Rh, float64 *deltaXp, float64 *X0p, float64 *sigmap, LVBoolean *busyp, int32 *iterationp);

CIN MgErr CINRun(TD1Hdl xh, TD1Hdl yh, TD1Hdl ph, TD1Hdl Xh, TD1Hdl Yh, TD1Hdl Rh, float64 *deltaXp, float64 *X0p, float64 *sigmap, LVBoolean *busyp, int32 *iterationp)
{
  float64 *x = (*xh)->data;	/* The x coordinates of the noisy data points, x[i+1] > x[i] */
  float64 *y = (*yh)->data;	/* The y coordinates of the noisy data points {(x[i],y[i])} */
  float64 *p;			/* The splinal interpolates at the data points x */

  float64 *X;			/* Spline joints X partitioning */
  float64 *Y;			/* Spline Y values to be fit in gliding FITDIMS chunks */
  float64 *R;			/* Second derivatives for gliding and final storage */

  int32 joint, point;

  /* Below follows the static storage that should service fitting iterations */
  static int32 joints, points;
  static int32 trail, lead, substart;
  static int32 precomp;			/* Indexes the A, AA, B and BB arrays during precomputation */
  static float64 deri, aTOL, gTOL;
  static float64 D;			/* The first derivative at the previously retired joint */
  static float64 Ytmpplus[FITDIMS+1];	/* Extended temporary storage; 0th element is for clamped point */

  /* Assumed to remain busy until proven otherwize */
  *busyp = LVTRUE;

  /* If these are the first few iterations, initialize and allocate some state */
  if ( *iterationp < 5)
    {
      /* During the first iteration initialize everything apart from some large allocations */
      if ( *iterationp == 0 )
        {
          float64 deltaX = *deltaXp;
          float64 X0 = *X0p;
          float64 miny, maxy;

          /* Cache and check the number of data points */
          points = (*xh)->dimSize;
          if (points < FITDIMS*4)
            {
              DbgPrintf("spline.c: too few data points. Not possible to fit a spline!\n");
              return bogusError;
            }

          /* Check if the x values increase monotonously */
          for (point = 1; point < points; point++)
            {
              if (x[point-1] >= x[point])
                {
                  DbgPrintf("spline.c: x values do not increase monotonously!\n");
                  return bogusError;
                }
            }

          /* Check of the x and y arrays have equal size */
          if ( points != (*yh)->dimSize )
            {
              DbgPrintf("spline.c: The y array must have the same length as the x array!\n");
              return bogusError;
            }

          /* Resize the p array to match x,y */
          if (noErr != NumericArrayResize(fD, 1, &(UHandle)ph, points))
            {
              DbgPrintf("spline.c: out of memory!");
              return bogusError;
            }
          (*ph)->dimSize = points;
     
          /* If the provided X array is empty compute it from the X0 and deltaX values plus the x range */
          if ( (*Xh)->dimSize == 0 )
            {
              float64 Xsum;

              /* Check if deltaX > 0 */
              if (deltaX <= 0)
                {
                  DbgPrintf("spline.c: deltaX must be positive and nonzero!");
                  return bogusError;
                }

              joints = 0.33+1+(x[point-1]-X0)/deltaX;

              /* Resize the X array to accomodate the joints */
              if (noErr != NumericArrayResize(fD, 1, &(UHandle)Xh, joints))
                {
                  DbgPrintf("spline.c: out of memory!");
                  return bogusError;
                }
              (*Xh)->dimSize = joints;

              /* Initialize the X positions of the joints  */
              Xsum = X0;
              for (joint = 0; joint < joints; joint++)
                {
                  (*Xh)->data[joint] = Xsum;
                  Xsum += deltaX;
                }
            }
          else
            {
              /* The user provided a set of custom joints */
              joints = (*Xh)->dimSize;

              /* Check if the custom joints increase monotonously */
              for (joint = 1; joint < joints; joint++)
                {
                  if ((*Xh)->data[joint-1] >= (*Xh)->data[joint])
                    {
                      DbgPrintf("spline.c: custom spline joint positions X do not increase monotonously!");
                      return bogusError;
                    }
                }
            }

          /* Check if the number of joints is reasonable */
          if ( joints < FITDIMS )
            {
              DbgPrintf("spline.c: too few spline joints, fit will not work!\n");
              return bogusError;
            }
          if ( joints > points/4 )
            {
              DbgPrintf("spline.c: too few points per joint < 4, fit will not be smooth!\n");
              return bogusError;
            }

          /* Reallocate the Y and R to the size of the X array */
          if ( ( noErr != NumericArrayResize(fD, 1, &(UHandle)Yh, joints) ) || ( noErr != NumericArrayResize(fD, 1, &(UHandle)Rh, joints) ) )
            {
              DbgPrintf("spline.c: out of memory!");
              return bogusError;
            }
          (*Yh)->dimSize = joints;
          (*Rh)->dimSize = joints;

            /* Determine the scale of variance so as to get a proper TOL and derivative */
            miny = y[0]; maxy = y[0];
            for (point = 1; point < points; point++)
              {
                if (y[point] < miny)  miny = y[point];
                if (y[point] > maxy)  maxy = y[point];
              }
            aTOL = 0.5e-6;
            gTOL = (maxy-miny)*1e-6;
            gTOL = gTOL*gTOL;
            deri = (maxy-miny)*1e-7;

            /* Initialize some of the cross call counts and indices */
            substart = 0;
            lead = 0;
            trail = 0;
            precomp = 0;

            /* Allocate work arrays for the spline() and interpolate() functions */
            alpha = (float64 *)DSNewPClr(joints*sizeof(float64));
            h = (float64 *)DSNewPClr(joints*sizeof(float64));
            l = (float64 *)DSNewPClr(joints*sizeof(float64));
            mu = (float64 *)DSNewPClr(joints*sizeof(float64));
            tri = (float64 *)DSNewPClr(joints*sizeof(float64));
        }
      /* Do the potentially huge allocations during seperate CIN call iterations */
      else if ( *iterationp == 1 )
        A = (float64 *)DSNewPClr(points*sizeof(float64));
      else if ( *iterationp == 2 )
         B = (float64 *)DSNewPClr(points*sizeof(float64));
      else if ( *iterationp == 3 )
         AA = (float64 *)DSNewPClr(points*sizeof(float64));
      else if ( *iterationp == 4 )
        {
          BB = (float64 *)DSNewPClr(points*sizeof(float64));
           if (alpha == NULL || h == NULL || l == NULL || mu == NULL || tri == NULL || A == NULL || B == NULL || AA == NULL || BB == NULL)
             {
               DbgPrintf("spline.c: out of memory!");
               return bogusError;
             }
         }
    }
  /* When all main initialization done... */
  else
    {
      /* Dereference some handles that might have changed during the initialization */
      p = (*ph)->data;
      X = (*Xh)->data;
      Y = (*Yh)->data;
      R = (*Rh)->data;

      /*
       * The subsequent bit is executed for each call of the CIN.
       * The first few calls break down the precomputation of the arrays needed for efficient
       * interpolation. The remaining calls break down the spline fit. Since steepest descent
       * fitting in more than six or so dimensions is prohibitively expensive we do not fit
       * using the entire spline.  Instead we start at the left of the x range and fit a small
       * spline with only a few joints, clamp the leftmost point of the fit, step one joint to
       * the right and iterate. Each CIN call executes one such a step.
       */
      if (precomp < points)
        {
          int32 j = 0;
          int32 k = 1;
          int32 chunkcount = 1000;		/* Precompute for 1000 points at a time */
          float64 ha = X[k]-X[j];

          for ( ; (precomp < points) && (chunkcount > 0); precomp++, chunkcount--)
            {
              /* Search for the spline covering the regions within which x[precomp] falls */
              while (x[precomp] > X[k] && k < joints-1)
                {
                  j += 1;
                  k += 1;
                  ha = X[k]-X[j];
                }

              /* The precomputed values below allow efficientl and stable interpolation */
              A[precomp] = (x[precomp]-X[j])/ha;
              B[precomp] = 1-A[precomp];
              AA[precomp] = (A[precomp]*A[precomp]*A[precomp]-A[precomp])*ha*ha/6;
              BB[precomp] = (B[precomp]*B[precomp]*B[precomp]-B[precomp])*ha*ha/6;
            }
        }
      else
        {
          /*
           * A steepest descent algorithm that tries to minimize minimize() for the leading FITDIMS
           * part spline.
           *
           * This algorithm was copied from "Numerical Analysis", Richard L. Burden & J. Douglas Faires,
           * 4th ed. The partial derivative evaluation is numerical.
           */
          int32 k;
          float64 a0, a2, a3, g, g0, g1, g2, g3, h1, h2, h3, z0;
          static float64 z[FITDIMS];
          float64 *Xc = &X[trail];
          float64 *Yc = &Y[trail];
          float64 *Ytmp = &Ytmpplus[1];

          /* Compute an initial approximation for the leading Y value via linear regression to the x points around its X */
          while (lead < trail+FITDIMS)
            {
              int32 subend;	/* One point past the end of the regression fit range for Y[lead] */

              while (x[substart] <  (X[lead] + X[max(0, lead-1)])/2 && substart < points-3)  substart += 1;
              subend = substart+3;
              while (x[subend] < (X[lead] + X[min(lead+1, joints-1)])/2 && subend < points)  subend += 1;
              Y[lead]  = regres(&x[substart], &y[substart], X[lead] , subend-substart);
              lead += 1;
            }

          /* The main loop; progressively closer after each descent step */
          k = 0;
          a3 = 1.0/64.0;			/* Decreases smoothly as the while loop converges */
          /* Determine the initial minimize(Y) */
          g = minimize(x, y, p, points, Xc, Yc, D, trail);
          while (k < MAXITERATIONS)
            {
              int32 j;
              float64 ahi, alo, aha;
              float64 ghi, glo, gha;

              /* Store the current minimize(Y) */
              g1 = g;

              /* Determine the partial derivative z and its norm */
              for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j];
              for (j = 0; j < FITDIMS; j++)
                {
                   Ytmp[j] += deri;
                   z[j] = (minimize(x, y, p, points, Xc, Ytmp, D, trail)-g1)/deri;
                   Ytmp[j] = Yc[j];
                }
              z0 = 0;
              for (j = 0; j < FITDIMS; j++)  z0 += z[j]*z[j];
              z0 = sqrt(z0);

              /* Normalize z, initialize some other shit as well */
              for (j = 0; j < FITDIMS; j++)  z[j] /= z0;

              /*
               * Try until within a factor of 2 of the largest step that makes an improvement.
               * This optimized algorithm minimizes the number of minimize() evaluations that
               * need to be made given the fact that from descent step to descent step the
               * fraction-of-the-normalized-derivative-that-can-at-most-be-moved-such-that
               * that-an-improvement-is-made varies gradually. Thus we start near the value
               * of the previous iteration.
               * Also, for the paraboloid extrapolation an evaluation at half this fraction
               * is needed. Thus the search is set up such that this value is remembered when
               * it happended to already have been evaluated during the search. The fact that
               * this value is needed anyway allows the search to move in larger steps, 1/4
               * instead of 1/2, without making superfluous calculations.
               */
              ahi = a3/2;
              for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - ahi*z[j];
              ghi = minimize(x, y, p, points, Xc, Ytmp, D, trail);
              if (ghi < g1)
                {
                  /* We are already within improvement range, increase to find bound */
                  alo = ahi; glo = ghi;
                  ahi *= 2;
                  for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - ahi*z[j];
                  ghi = minimize(x, y, p, points, Xc, Ytmp, D, trail);
                  if (ghi < g1)
                    {
                      /* We are still in the improvement range, increase to find bound until found */
                      do
                        {
                          aha = alo; gha = glo; alo = ahi; glo = ghi;
                          ahi *= 2;
                          for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - ahi*z[j];
                          ghi = minimize(x, y, p, points, Xc, Ytmp, D, trail);
                        }
                      while (ghi < g1);
                    }
                  else
                    {
                      /* We found the bound, calculate the half value and terminate */
                      aha = alo/2;
                      for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - aha*z[j];
                      gha = minimize(x, y, p, points, Xc, Ytmp, D, trail);
                    }
                }
              else
                {
                  /* We are still outside improvement range, decrease at double speed until within range */
                  aha = ahi; gha = ghi;
                  do
                    {
                      aha /= 4;
                      for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - aha*z[j];
                      gha = minimize(x, y, p, points, Xc, Ytmp, D, trail);
                    }
                  while (gha >= g1 && aha > aTOL);

                  /* If further improvement unlikely, break out of the loop */
                  if (aha <= aTOL)
                    break;

                  /* Evaluate the skipped thingy */
                  alo = aha*2;
                  for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - alo*z[j];
                  glo = minimize(x, y, p, points, Xc, Ytmp, D, trail);
                  if (glo < g1)
                    {
                      /* We got it */ 
                    }
                  else
                    {
                      /* Step one further down */
                      alo = aha; glo = gha;
                      aha = alo/2;
                      for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - aha*z[j];
                      gha = minimize(x, y, p, points, Xc, Ytmp, D, trail);
                    }
                }
              a3 = alo; a2 = aha;
              g3 = glo; g2 = gha;

              /* Extrapolate to parabola minimum (all smooth minima can be fit with a parabola up to 2nd order) */
              h1 = (g2-g1)/a2;
              h2 = (g3-g2)/(a3-a2);
              h3 = (h2-h1)/a3;
              a0 = 0.5*(a2-h1/h3);
              for (j = 0; j < FITDIMS; j++)  Ytmp[j] = Yc[j] - a0*z[j];
              g0 = minimize(x, y, p, points, Xc, Ytmp, D, trail);

              /* Choose the largest improvement and adjust the fit parameters Yc[] accordingly */
              if (g0 < g3 && g0 < g2)
                {
                  /* The extrapolate parabola trick worked, we must be near the minimum */
                  for (j = 0; j < FITDIMS; j++)  Yc[j] -= a0*z[j];
                  g = g0;
                }
              else if (g2 < g3)
                { 
                  for (j = 0; j < FITDIMS; j++)  Yc[j] -= a2*z[j];
                  g = g2;
                }
              else
                { 
                  for (j = 0; j < FITDIMS; j++)  Yc[j] -= a3*z[j];
                  g = g3;
                }

              if ( fabs(g1-g) < gTOL )
                break;

              k += 1;
            }
          if (k == MAXITERATIONS)
            {
              DbgPrintf("spline.c: maximum iterations exceeded without convergence within %le!\n", aTOL);
              return bogusError;
            }

          /* The leading spline has been fit, recalculate the spline */
          if (trail == 0)
            spline(R, X, Y, NATURAL, NATURAL, FITDIMS);
          else
            spline(&R[trail-1], &X[trail-1], &Y[trail-1], D, NATURAL, FITDIMS+1);

          /* Use the recalculated spline to determine the derivative of the trailing joint for clamping */
          D = (Y[trail+1]-Y[trail])/(X[trail+1]-X[trail])-(X[trail+1]-X[trail])*R[trail]/3-(X[trail+1]-X[trail])*R[trail+1]/6;

          /* Also fix the Y value in the extended temporary array */
          Ytmpplus[0] = Y[trail];

          /* Was this the last iteration? */
          if (trail < joints-FITDIMS)
            {
              /* We're not done yet, bump the trailing joint index */
              trail += 1;
            }
          else
            {
              float64 sigmasum;

              /* We're done, calculate the entire spline and the x interpolate */
              spline(R, X, Y, NATURAL, NATURAL, joints);

              interpolate(p, x, 0, points-1, X, Y, R, joints);
              sigmasum = 0;
              for (point = 0; point < points; point++)
                {
                  sigmasum += (p[point]-y[point])*(p[point]-y[point]);
                }
              *sigmap = sqrt(sigmasum/points);

              /* Deallocate the spline() function related storage */
              DisposeIfNN(alpha);
              DisposeIfNN(h);
              DisposeIfNN(l);
              DisposeIfNN(mu);
              DisposeIfNN(tri);
              DisposeIfNN(A);
              DisposeIfNN(B);
              DisposeIfNN(AA);
              DisposeIfNN(BB);

              /* Signal calling loop to break */
              *busyp = LVFALSE;
            }
        }
    }

  return noErr;
}



/*** SUPPORT FUNCTIONS ***/

/*
 * This evaluates the function to be minimized;
 * It determines the FITDIMS or FITDIMS+1 joint spline for the specified X and Y values
 * and computes the average squared differences between the y values and the corresponding
 * spline interpolates in the range covered by the spline joints.
 * When trail > 0, X and Y are bumped down by one so as to include the clamped point whose
 * derivative must be specified via the D parameter. In this case the spline has FITDIMS+1
 * joints.
 */
float64
minimize(float64 *x, float64 *y, float64 *p, int32 points, float64 *X, float64 *Y, float64 D, int32 trail)
{
  float64 sum, left;
  int32 point, firstpoint, lastpoint, lo, mid, hi, pjoints;
  static float64 R[FITDIMS+1];

  if (trail == 0)
    {
      /* If no previous fit, the spline has as joints those whose Y values will be fit */
      pjoints = FITDIMS;
      left = NATURAL;
    }
  else
    {
      /* If a previous fit the leftmost joint will be clamped, the others will be fit */
      pjoints = FITDIMS+1;
      left = D;
      X -= 1;
      Y -= 1;
    }
   
  /* Find the first x value in the spline's range using a binary search */
  lo = 0;
  hi = points-1;
  while (lo != hi)
    {
      mid = ( (lo + hi) >> 1 );
      if (x[mid] < X[0])
        lo = mid+1;
      else
        hi = mid;
    }
  firstpoint = lo;

  /* Find the last x value in the spline's range using a binary search */
  lo = 0;
  hi = points-1;
  while (hi != lo)
    {
      mid = ( (lo + hi + 1) >> 1 );
      if (x[mid] > X[pjoints-1])
        hi = mid-1;
      else
        lo = mid;
    }
  lastpoint = hi;

  /* Obtain the splinal piece */
  spline(R, X, Y, left, NATURAL, pjoints);

  /* Interpolate the covered point range */
  interpolate(p, x, firstpoint, lastpoint, X, Y, R, pjoints);

  /*
   * Sum the squared differences and return the normalized sum. This is a fairly expensive bit.
   * Break it down such that the pipelines can keep on flowing.
   */
  sum = 0;
  for (point = firstpoint; point <= lastpoint; point++)
    p[point] -= y[point];
  for (point = firstpoint; point <= lastpoint; point++)
    p[point] *= p[point];
  for (point = firstpoint; point <= lastpoint; point++)
    sum += p[point];

  return sum/(lastpoint-firstpoint+1);
}

/*
 * When a boundary conditions are set to 10^30 it will be taken to be natural, that is,
 * the second derivative will be forced to 0
 *
 * This algorithm was obtained by combining the two algorithms pertaining to cubic
 * splines in "Numerical Analysis", Richard L. Burden & J. Douglas Faires, 4th ed.
 */
void
spline(float64 *R, float64 *X, float64 *Y, float64 left, float64 right, int32 joints)
{
  /*
   * The global float64 arrays called alpha, h, l, mu, tri each are initialized
   * joints elements.
   */
  int32 i, n;

  n = joints-1;

  /* Initialize the x range values */
  for (i = 0; i < n; i++)  h[i] = X[i+1] - X[i];

  /* Handle clamped boundary related initialization */
  if (left != NATURAL)
    alpha[0] = 3*(Y[1]-Y[0])/h[0] - 3*left;
  if (right != NATURAL)
    alpha[n] = 3*right - 3*(Y[n]- Y[n-1])/h[n-1];

  /* Set up the remainder of the alpha array */
  for (i = 1; i < n; i++)  alpha[i] = 3*(Y[i+1]*h[i-1] - Y[i]*(X[i+1]-X[i-1]) + Y[i-1]*h[i])/(h[i-1]*h[i]);

  /* Fumble one way or another depending on whether the left boundary is clamped */
  if (left != NATURAL)
    {
      /* Clamped */
      l[0] = 2*h[0];
      mu[0] = 0.5;
      tri[0] = alpha[0]/l[0];
    }
  else
    {
      /* Natural */
      l[0] = 1;
      mu[0] = 0;
      tri[0] = 0;
    }

  /* Solve what amounts to being a tridiagonal system */
  for (i = 1; i < n; i++)
    {
      l[i] = 2*(X[i+1] - X[i-1])- h[i-1]*mu[i-1];
      mu[i] = h[i]/l[i];
      tri[i] = (alpha[i] - h[i-1]*tri[i-1])/l[i];
    }

  /* Fumble one way or another depending on whether the left boundary is clamped */
  if (right != NATURAL)
    {
      /* Clamped */
      l[n] = h[n-1]*(2-mu[n-1]);
      tri[n] = (alpha[n] - h[n-1]*tri[n-1])/l[n];
      R[n] = tri[n];
    }
  else
    {
      /* Natural */
      l[n] = 1;
      tri[n] = 0;
      R[n] = 0;
    }

  /* Calculate the resulting second derivatives which allow convenient interpolation */
  for (i = n-1; i >= 0; i--)  R[i] = tri[i] - mu[i]*R[i+1];
  for (i = 0; i <= n; i++)  R[i] *= 2;
}

/*
 * Given the X, Y and second derivative arrays for a spline, calculate the interpolation
 * at monotonously increasing set of points x and put the result into an array p;
 */
void
interpolate(float64 *p, float64 *x, int32 firstpoint, int32 lastpoint, float64 *X, float64 *Y, float64 *R, int32 joints)
{
  int32 point;
  int32 j = 0;
  int32 k = 1;

  for (point = firstpoint; point <= lastpoint; point++)
    {
      /* Search for the spline covering the regions within which x[point] falls */
      while (x[point] > X[k] && k < joints-1)
        {
          j += 1;
          k += 1;
        }
      /* Interpolate (before X[0] or beyond X[joints-1] extrapolate) */
      p[point] = B[point]*Y[j]+A[point]*Y[k]+BB[point]*R[j]+AA[point]*R[k];
    }
}

/*
 * This function performs a regression fit to a number of points and determines the
 * y intercept at a specified X position.
 */
float64
regres(float64 *xval, float64 *yval, float64 X, int32 elements)
{
  int32 index;
  float64 xave, yave, sumsqdifx, slope;

  /* Determine the centroid */
  xave = 0; yave = 0;
  for (index = 0; index < elements; index++)
    {
      xave += xval[index];
      yave += yval[index];
    }
  xave /= elements; yave /= elements;

  /* Determine the sum of squared x differences */
  sumsqdifx = 0;
  for(index = 0; index < elements; index++)  sumsqdifx += (xval[index]-xave)*(xval[index]-xave);

  /* Determine the slope */
  slope = 0;
  for(index = 0; index < elements; index++)  slope += (xval[index]-xave)*(yval[index]-yave);
  slope /= sumsqdifx;

  /* Return the intercept at the specified X position */
  return yave+slope*(X-xave);
}

void
DisposeIfNN(float64 *allocedptr)
{
  if (allocedptr != NULL)
    {
      DSDisposePtr(allocedptr);
    }
}
